==> torch-inspect-0.0.1.tar.gz: torch-inspect-0.0.1/torch_inspect/inspect.py <==
import sys
from dataclasses import dataclass
from functools import partial
from typing import List, IO, Union, Type

import numpy as np
import torch
import torch.nn as nn

from torch.utils.hooks import RemovableHandle


__all__ = ('LayerInfo', 'NetworkInfo', 'inspect', 'summary')


CPU = 'cpu'
CUDA = 'cuda'


@dataclass
class LayerInfo:
    name: str
    input_shape: List[int]
    output_shape: List[int]
    trainable: bool
    nb_params: int


@dataclass
class NetworkInfo:
    total_params: int
    trainable_params: int
    total_input_size: int
    total_output_size: int
    total_params_size: int
    total_size: int


def make_network_info(
    info_list: List[LayerInfo], input_size: List[int], batch_size: int = -1
) -> NetworkInfo:
    trainable_params = 0
    total_params = 0
    total_output = 0
    for layer in info_list:
        total_params += layer.nb_params
        total_output += abs(np.prod(layer.output_shape))
        if layer.trainable:
            trainable_params += layer.nb_params

    # assume 4 bytes/number (float on cuda).
    total_input_size = np.prod(input_size) * abs(batch_size) * 4
    # x2 for gradients
    total_output_size = 2 * total_output * 4
    total_params_size = total_params * 4
    total_size = total_params_size + total_output_size + total_input_size
    ni = NetworkInfo(
        total_params,
        trainable_params,
        total_input_size,
        total_output_size,
        total_params_size,
        total_size,
    )
    return ni


def should_attach_hook(model: nn.Module, module: nn.Module) -> bool:
    v = (
        not isinstance(module, nn.Sequential)
        and not isinstance(module, nn.ModuleList)
        and not (module is model)  # exclude self
    )
    return v


_FTensorType = Type[Union[torch.cuda.FloatTensor, torch.FloatTensor]]


def _infer_dtype(device: str) -> _FTensorType:
    device = device.lower()
    if device not in (CUDA, CPU):
        msg = 'Input device is not valid, please specify "cuda" or "cpu"'
        raise ValueError(msg)

    if device == CUDA and torch.cuda.is_available():
        dtype = torch.cuda.FloatTensor
    else:
        dtype = torch.FloatTensor
    return dtype


def _has_weight(module: nn.Module) -> bool:
    return hasattr(module, 'weight') and hasattr(module.weight, 'size')


def _has_bias(module: nn.Module) -> bool:
    return hasattr(module, 'weight') and hasattr(module.weight, 'size')


class _ModuleHook:
    def __init__(self, batch_size: int):
        self.batch_size = batch_size
        self.layer_list: List[LayerInfo] = []

    def hook(
        self, module: nn.Module, input: _FTensorType, output: _FTensorType
    ) -> None:

        # make layer name
        module_idx = len(self.layer_list)
        class_name = str(module.__class__).split('.')[-1].split("'")[0]
        name = f'{class_name}-{module_idx + 1}'

        # calculate input shape
        input_shape = [self.batch_size] + list(input[0].size())[1:]

        # calculate output shape
        if isinstance(output, (list, tuple)):
            output_shape = [[-1] + list(o.size())[1:] for o in output]
        else:
            output_shape = [self.batch_size] + list(output.size())[1:]

        # calculate number of params
        params = 0
        trainable = False

        if _has_weight(module):
            params += np.prod(module.weight.size())
            trainable = module.weight.requires_grad

        if _has_bias(module):
            params += np.prod(module.bias.size())

        # recored result
        layer = LayerInfo(
            name, input_shape, output_shape, trainable, int(params)
        )
        self.layer_list.append(layer)


def inspect(
    model: nn.Module,
    input_size: List[int],
    batch_size: int = -1,
    device: str = CPU,
) -> List[LayerInfo]:
    hook = _ModuleHook(batch_size)
    handles: List[RemovableHandle] = []

    def register_hook(module: nn.Module) -> None:
        if should_attach_hook(model, module):
            handle: RemovableHandle = module.register_forward_hook(hook.hook)
            handles.append(handle)

    dtype = _infer_dtype(device)

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # make fake input with batch_size of 2 for batchnorm
    x = [torch.rand(2, *in_size).type(dtype) for in_size in input_size]

    # attach hooks to each applicable layer
    model.apply(register_hook)

    # forward pass
    try:
        model(*x)
    finally:
        # cleanup all attached hooks, to move model to original state
        for h in handles:
            h.remove()

    return hook.layer_list


def summary(
    model: nn.Module,
    input_size: List[int],
    batch_size: int = -1,
    device: str = CPU,
    file: IO[str] = sys.stdout,
    flush: bool = False,
) -> None:
    summary = inspect(model, input_size, batch_size=batch_size, device=device)
    n = make_network_info(summary, input_size, batch_size)
    print_ = partial(print, file=file, flush=flush)
    print_('\n')
    print_('----------------------------------------------------------------')
    line_new = '{:>20}  {:>25} {:>15}'.format(
        'Layer (type)', 'Output Shape', 'Param #'
    )
    print_(line_new)
    print_('================================================================')

    for layer in summary:
        line_new = '{:>20}  {:>25} {:>15}'.format(
            layer.name,
            str(layer.output_shape),
            '{0:,}'.format(layer.nb_params),
        )
        print_(line_new)
    MB = 1024 ** 2
    print_('================================================================')
    print_(f'Total params: {n.total_params:,}')
    print_(f'Trainable params: {n.trainable_params:,}')
    print_(f'Non-trainable params: {n.total_params - n.trainable_params:,}')
    print_('----------------------------------------------------------------')
    print_(f'Input size (MB): {n.total_input_size / MB:0.2f}')
    print_(f'Forward/backward pass size (MB): {n.total_output_size / MB:0.2f}')
    print_(f'Params size (MB): {n.total_params_size / MB :0.2f}')
    print_(f'Estimated Total Size (MB): {n.total_size / MB:0.2f}')
    print_('----------------------------------------------------------------')

==> torch-inspect-0.0.2.tar.gz: torch-inspect-0.0.2/torch_inspect/inspect.py <==
import sys
from dataclasses import dataclass
from functools import partial
from typing import List, IO, Union, Tuple

import numpy as np
import torch
import torch.nn as nn

from torch.utils.hooks import RemovableHandle


__all__ = ('LayerInfo', 'NetworkInfo', 'inspect', 'summary')


InputShape = Tuple[int, ...]


@dataclass
class LayerInfo:
    name: str
    input_shape: List[int]
    output_shape: List[int]
    trainable_params: int
    non_trainable_params: int


@dataclass
class NetworkInfo:
    total_params: int
    trainable_params: int
    total_input_size: int
    total_output_size: int
    total_params_size: int
    total_size: int


def make_network_info(
    info_list: List[LayerInfo],
    input_size: Union[InputShape, List[InputShape]],
    batch_size: int = -1,
) -> NetworkInfo:
    trainable_params = 0
    total_params = 0
    total_output = 0
    for layer in info_list:
        total_params += layer.trainable_params + layer.non_trainable_params
        total_output += abs(np.prod(layer.output_shape))
        trainable_params += layer.trainable_params

    # assume 4 bytes/number (float on cuda).
    total_input_size = np.prod(input_size) * abs(batch_size) * 4
    # x2 for gradients
    total_output_size = 2 * total_output * 4
    total_params_size = total_params * 4
    total_size = total_params_size + total_output_size + total_input_size
    ni = NetworkInfo(
        total_params,
        trainable_params,
        total_input_size,
        total_output_size,
        total_params_size,
        total_size,
    )
    return ni


def should_attach_hook(model: nn.Module, module: nn.Module) -> bool:
    v = (
        not isinstance(module, nn.Sequential)
        and not isinstance(module, nn.ModuleList)
        and not (module is model)  # exclude self
    )
    return v


def _has_weight(module: nn.Module) -> bool:
    return hasattr(module, 'weight') and hasattr(module.weight, 'size')


def _has_bias(module: nn.Module) -> bool:
    return hasattr(module, 'bias') and hasattr(module.bias, 'size')


def _has_running_mean(module: nn.Module) -> bool:
    return (
        hasattr(module, 'running_mean')
        and hasattr(module.running_mean, 'size')
        and hasattr(module, 'track_running_stats')
        and module.track_running_stats
    )


def _has_running_var(module: nn.Module) -> bool:
    return (
        hasattr(module, 'running_var')
        and hasattr(module.running_var, 'size')
        and hasattr(module, 'track_running_stats')
        and module.track_running_stats
    )


class _ModuleHook:
    def __init__(self, batch_size: int):
        self.batch_size = batch_size
        self.layer_list: List[LayerInfo] = []

    def hook(
        self,
        module: nn.Module,
        input: torch.FloatTensor,
        output: torch.FloatTensor,
    ) -> None:

        # make layer name
        module_idx = len(self.layer_list)
        class_name = str(module.__class__).split('.')[-1].split("'")[0]
        name = f'{class_name}-{module_idx + 1}'

        # calculate input shape
        input_shape = [self.batch_size] + list(input[0].size())[1:]

        # calculate output shape
        if isinstance(output, (list, tuple)):
            output_shape = [[-1] + list(o.size())[1:] for o in output]
        else:
            output_shape = [self.batch_size] + list(output.size())[1:]

        # calculate number of params
        trainable_params = 0
        non_trainable_params = 0

        if _has_weight(module):
            params = np.prod(module.weight.size())
            if module.weight.requires_grad:
                trainable_params += params
            else:
                non_trainable_params += params

        if _has_bias(module):
            params = np.prod(module.bias.size())
            if module.bias.requires_grad:
                trainable_params += params
            else:
                non_trainable_params += params

        if _has_running_mean(module):
            params = np.prod(module.running_mean.size())
            non_trainable_params += params

        if _has_running_var(module):
            params = np.prod(module.running_var.size())
            non_trainable_params += params

        # recored result
        layer = LayerInfo(
            name,
            input_shape,
            output_shape,
            int(trainable_params),
            int(non_trainable_params),
        )
        self.layer_list.append(layer)


def inspect(
    model: nn.Module,
    input_size: Union[InputShape, List[InputShape]],
    batch_size: int = -1,
) -> List[LayerInfo]:
    hook = _ModuleHook(batch_size)
    handles: List[RemovableHandle] = []

    def register_hook(module: nn.Module) -> None:
        if should_attach_hook(model, module):
            h: RemovableHandle = module.register_forward_hook(  # type: ignore
                hook.hook
            )
            handles.append(h)

    dtype = torch.FloatTensor

    # multiple inputs to the network
    if isinstance(input_size, tuple):
        input_size = [input_size]

    # make fake input with batch_size of 2 for batchnorm
    x = [
        torch.rand(2, *in_size).type(dtype)  # type: ignore
        for in_size in input_size
    ]

    # attach hooks to each applicable layer
    model.apply(register_hook)  # type: ignore

    # forward pass
    try:
        model(*x)
    finally:
        # cleanup all attached hooks, to move model to original state
        for h in handles:
            h.remove()  # type: ignore

    return hook.layer_list


def summary(
    model: nn.Module,
    input_size: Union[InputShape, List[InputShape]],
    batch_size: int = -1,
    file: IO[str] = sys.stdout,
    flush: bool = False,
) -> NetworkInfo:
    summary = inspect(model, input_size, batch_size=batch_size)
    n = make_network_info(summary, input_size, batch_size)
    print_ = partial(print, file=file, flush=flush)
    print_('\n')
    print_('----------------------------------------------------------------')
    line_new = '{:>20}  {:>25} {:>15}'.format(
        'Layer (type)', 'Output Shape', 'Param #'
    )
    print_(line_new)
    print_('================================================================')

    for layer in summary:
        total_params = layer.trainable_params + layer.non_trainable_params
        line_new = '{:>20}  {:>25} {:>15}'.format(
            layer.name, str(layer.output_shape), '{0:,}'.format(total_params)
        )
        print_(line_new)
    MB = 1024 ** 2
    print_('================================================================')
    print_(f'Total params: {n.total_params:,}')
    print_(f'Trainable params: {n.trainable_params:,}')
    print_(f'Non-trainable params: {n.total_params - n.trainable_params:,}')
    print_('----------------------------------------------------------------')
    print_(f'Input size (MB): {n.total_input_size / MB:0.2f}')
    print_(f'Forward/backward pass size (MB): {n.total_output_size / MB:0.2f}')
    print_(f'Params size (MB): {n.total_params_size / MB :0.2f}')
    print_(f'Estimated Total Size (MB): {n.total_size / MB:0.2f}')
    print_('----------------------------------------------------------------')
    return n
